@@ -14,7 +14,4 @@
 #undef HAVE_TIMEZONE
 #undef HAVE_TM_GMTOFF
 
-/* build with extra features? */
-#undef USE_GTKHTML
-
 #undef  GNOME_ICONDIR
