@@ -20,24 +20,50 @@
  *
  */
 
-#include "gmime-message.h"
-#include "gmime-utils.h"
-#include "internet-address.h"
+#ifdef HAVE_CONFIG_H
 #include <config.h>
+#endif
+
 #include <string.h>
 #include <ctype.h>
 #include <locale.h>
 
+#include "gmime-message.h"
+#include "gmime-utils.h"
+#include "internet-address.h"
+#include "gmime-stream-mem.h"
+
+
+static char *rfc822_headers[] = {
+	"Return-Path",
+	"Received",
+	"Date",
+	"From",
+	"Reply-To",
+	"Subject",
+	"Sender",
+	"To",
+	"Cc",
+	NULL
+};
+
+
 /**
  * g_mime_message_new: Create a new MIME Message object
+ * @init_headers: initialize the headers 
+ *
+ * If @init_headers is %TRUE, then the standard rfc822 headers are
+ * initialized so as to put headers in a nice friendly order.
  *
  * Returns an empty MIME Message object with no headers nor content
  * set by default.
  **/
 GMimeMessage *
-g_mime_message_new ()
+g_mime_message_new (gboolean init_headers)
 {
 	GMimeMessage *message;
+	GMimeHeader *headers;
+	int i;
 	
 	message = g_new0 (GMimeMessage, 1);
 	
@@ -45,7 +71,13 @@ g_mime_message_new ()
 	
 	message->header->recipients = g_hash_table_new (g_str_hash, g_str_equal);
 	
-	message->header->arbitrary_headers = g_ptr_array_new ();
+	message->header->headers = headers = g_mime_header_new ();
+	
+	if (init_headers) {
+		/* Populate with the "standard" rfc822 headers so we can have a standard order */
+		for (i = 0; rfc822_headers[i]; i++) 
+			g_mime_header_set (headers, rfc822_headers[i], NULL);
+	}
 	
 	return message;
 }
@@ -81,8 +113,6 @@ recipients_destroy (gpointer key, gpointer value, gpointer user_data)
 void
 g_mime_message_destroy (GMimeMessage *message)
 {
-	int i;
-	
 	g_return_if_fail (message != NULL);
 	
 	g_free (message->header->from);
@@ -96,15 +126,7 @@ g_mime_message_destroy (GMimeMessage *message)
 	
 	g_free (message->header->message_id);
 	
-	/* destroy arbitrary headers */
-	for (i = 0; i < message->header->arbitrary_headers->len; i++) {
-		GMimeHeader *header = message->header->arbitrary_headers->pdata[i];
-		
-		g_free (header->name);
-		g_free (header->value);
-		g_free (header);
-	}
-	g_ptr_array_free (message->header->arbitrary_headers, TRUE);
+	g_mime_header_destroy (message->header->headers);
 	
 	g_free (message->header);
 	
@@ -127,12 +149,17 @@ g_mime_message_destroy (GMimeMessage *message)
 void
 g_mime_message_set_sender (GMimeMessage *message, const gchar *sender)
 {
+	GMimeMessageHeader *header;
+	
 	g_return_if_fail (message != NULL);
 	
-	if (message->header->from)
-		g_free (message->header->from);
+	header = message->header;
 	
-	message->header->from = g_strdup (sender);
+	if (header->from)
+		g_free (header->from);
+	
+	header->from = g_strstrip (g_strdup (sender));
+	g_mime_header_set (header->headers, "From", header->from);
 }
 
 
@@ -161,12 +188,17 @@ g_mime_message_get_sender (GMimeMessage *message)
 void
 g_mime_message_set_reply_to (GMimeMessage *message, const gchar *reply_to)
 {
+	GMimeMessageHeader *header;
+	
 	g_return_if_fail (message != NULL);
 	
-	if (message->header->reply_to)
-		g_free (message->header->reply_to);
+	header = message->header;
 	
-	message->header->reply_to = g_strdup (reply_to);
+	if (header->reply_to)
+		g_free (header->reply_to);
+	
+	header->reply_to = g_strstrip (g_strdup (reply_to));
+	g_mime_header_set (header->headers, "Reply-To", header->reply_to);
 }
 
 
@@ -185,6 +217,37 @@ g_mime_message_get_reply_to (GMimeMessage *message)
 }
 
 
+static void
+sync_recipient_header (GMimeMessage *message, const gchar *type)
+{
+	GList *recipients;
+	
+	/* sync the specified recipient header */
+	recipients = g_mime_message_get_recipients (message, type);
+	if (recipients) {
+		GString *recip;
+		
+		recip = g_string_new ("");
+		while (recipients) {
+			InternetAddress *ia;
+			gchar *address;
+			
+			ia = recipients->data;
+			address = internet_address_to_string (ia, TRUE);
+			g_string_append (recip, address);
+			g_free (address);
+			
+			recipients = recipients->next;
+			if (recipients)
+				g_string_append (recip, ", ");
+		}
+		
+		g_mime_header_set (message->header->headers, type, recip->str);
+		g_string_free (recip, TRUE);
+	} else
+		g_mime_header_set (message->header->headers, type, NULL);
+}
+
 /**
  * g_mime_message_add_recipient: Add a recipient
  * @message: MIME Message to change
@@ -202,13 +265,14 @@ g_mime_message_add_recipient (GMimeMessage *message, gchar *type, const gchar *n
 	InternetAddress *ia;
 	GList *recipients;
 	
-	ia = internet_address_new (name, address);
+	ia = internet_address_new_name (name, address);
 	
 	recipients = g_hash_table_lookup (message->header->recipients, type);
 	g_hash_table_remove (message->header->recipients, type);
 	
 	recipients = g_list_append (recipients, ia);
 	g_hash_table_insert (message->header->recipients, type, recipients);
+	sync_recipient_header (message, type);
 }
 
 
@@ -227,9 +291,7 @@ g_mime_message_add_recipient (GMimeMessage *message, gchar *type, const gchar *n
 void
 g_mime_message_add_recipients_from_string (GMimeMessage *message, gchar *type, const gchar *string)
 {
-	InternetAddress *ia;
-	GList *recipients;
-	gchar *ptr, *eptr, *recipient;
+	GList *recipients, *addrlist;
 	
 	g_return_if_fail (message != NULL);
 	g_return_if_fail (string != NULL);
@@ -237,40 +299,13 @@ g_mime_message_add_recipients_from_string (GMimeMessage *message, gchar *type, c
 	recipients = g_hash_table_lookup (message->header->recipients, type);
 	g_hash_table_remove (message->header->recipients, type);
 	
-	ptr = (gchar *) string;
-	
-	while (*ptr) {
-		gboolean in_quotes = FALSE;
-		
-		/* skip through leading whitespace */
-		for ( ; *ptr && isspace (*ptr); ptr++);
-		
-		if (*ptr == '"')
-			in_quotes = TRUE;
-		
-		/* find the end of this address */
-		eptr = ptr + 1;
-		while (*eptr) {
-			if (*eptr == '"' && *(eptr - 1) != '\\')
-				in_quotes = !in_quotes;
-			else if (*eptr == ',' && !in_quotes)
-				break;
-			
-			eptr++;
-		}
-		
-		recipient = g_strndup (ptr, (gint) (eptr - ptr));
-		ia = internet_address_new_from_string (recipient);
-		g_free (recipient);
-		recipients = g_list_append (recipients, ia);
-		
-		if (*eptr)
-			ptr = eptr + 1;
-		else
-			break;
-	}
+	addrlist = internet_address_parse_string (string);
+	if (addrlist)
+		recipients = g_list_concat (recipients, addrlist);
 	
 	g_hash_table_insert (message->header->recipients, type, recipients);
+	
+	sync_recipient_header (message, type);
 }
 
 
@@ -303,12 +338,17 @@ g_mime_message_get_recipients (GMimeMessage *message, const gchar *type)
 void
 g_mime_message_set_subject (GMimeMessage *message, const gchar *subject)
 {
+	GMimeMessageHeader *header;
+	
 	g_return_if_fail (message != NULL);
 	
-	if (message->header->subject)
-		g_free (message->header->subject);
+	header = message->header;
 	
-	message->header->subject = g_strdup (subject);
+	if (header->subject)
+		g_free (header->subject);
+	
+	header->subject = g_strstrip (g_strdup (subject));
+	g_mime_header_set (header->headers, "Subject", header->subject);
 }
 
 
@@ -338,10 +378,16 @@ g_mime_message_get_subject (GMimeMessage *message)
 void
 g_mime_message_set_date (GMimeMessage *message, time_t date, int gmt_offset)
 {
+	gchar *date_str;
+	
 	g_return_if_fail (message != NULL);
 	
 	message->header->date = date;
 	message->header->gmt_offset = gmt_offset;
+	
+	date_str = g_mime_message_get_date_string (message);
+	g_mime_header_set (message->header->headers, "Date", date_str);
+	g_free (date_str);
 }
 
 
@@ -402,11 +448,17 @@ g_mime_message_get_date_string (GMimeMessage *message)
 void
 g_mime_message_set_message_id (GMimeMessage *message, const gchar *id)
 {
+	GMimeMessageHeader *header;
+	
 	g_return_if_fail (message != NULL);
 	
-	if (message->header->message_id)
-		g_free (message->header->message_id);
-	message->header->message_id = g_strdup (id);
+	header = message->header;
+	
+	if (header->message_id)
+		g_free (header->message_id);
+	
+	header->message_id = g_strstrip (g_strdup (id));
+	g_mime_header_set (header->headers, "Message-Id", header->message_id);
 }
 
 
@@ -426,7 +478,7 @@ g_mime_message_get_message_id (GMimeMessage *message)
 
 
 /**
- * g_mime_message_add_arbitrary_header: Add an arbitrary message header
+ * g_mime_message_add_header: Add an arbitrary message header
  * @message: MIME Message
  * @field: rfc822 header field
  * @value: the contents of the header field
@@ -435,17 +487,45 @@ g_mime_message_get_message_id (GMimeMessage *message)
  * X-Priority, or In-Reply-To.
  **/
 void
-g_mime_message_add_arbitrary_header (GMimeMessage *message, const gchar *field, const gchar *value)
+g_mime_message_add_header (GMimeMessage *message, const gchar *field, const gchar *value)
 {
-	GMimeHeader *header;
-	
 	g_return_if_fail (message != NULL);
 	
-	header = g_new (GMimeHeader, 1);
-	header->name = g_strdup (field);
-	header->value = g_strdup (value);
+	g_mime_header_add (message->header->headers, field, value);
+}
+
+
+/**
+ * g_mime_message_set_header: Add an arbitrary message header
+ * @message: MIME Message
+ * @field: rfc822 header field
+ * @value: the contents of the header field
+ *
+ * Set an arbitrary message header to the MIME Message such as X-Mailer,
+ * X-Priority, or In-Reply-To.
+ **/
+void
+g_mime_message_set_header (GMimeMessage *message, const gchar *field, const gchar *value)
+{
+	g_return_if_fail (message != NULL);
 	
-	g_ptr_array_add (message->header->arbitrary_headers, header);
+	g_mime_header_set (message->header->headers, field, value);
+}
+
+
+/**
+ * g_mime_message_get_header:
+ * @message: MIME Message
+ * @field: rfc822 header field
+ *
+ * Returns the value of the requested header (or NULL if it isn't set)
+ **/
+const gchar *
+g_mime_message_get_header (GMimeMessage *message, const gchar *field)
+{
+	g_return_val_if_fail (message != NULL, NULL);
+	
+	return g_mime_header_get (message->header->headers, field);
 }
 
 
@@ -468,132 +548,26 @@ g_mime_message_set_mime_part (GMimeMessage *message, GMimePart *mime_part)
 }
 
 
-static gchar *
-create_header (GMimeMessage *message)
+/**
+ * g_mime_message_write_to_stream: Write the MIME Message to a stream
+ * @message: MIME Message
+ * @stream: output stream
+ *
+ * Write the contents of the MIME Message to @string.
+ **/
+void
+g_mime_message_write_to_stream (GMimeMessage *message, GMimeStream *stream)
 {
-	/* Make sure to not append "\n\n" to the end
-	 * so we can add MIME headers later */
-	GString *string;
-	gchar *str, *buf, *date, *subject;
-	GList *recipients;
+	g_return_if_fail (message != NULL);
+	g_return_if_fail (stream != NULL);
 	
-	string = g_string_new ("");
+	g_mime_header_write_to_stream (message->header->headers, stream);
 	
-	/* write out the arbitrary headers first (as they may contain
-	 * "Received:" headers which really should come first) */
-	if (message->header->arbitrary_headers->len) {
-		gint i;
-		
-		for (i = 0; i < message->header->arbitrary_headers->len; i++) {
-			const GMimeHeader *header;
-			gchar *encoded_value;
-			
-			header = message->header->arbitrary_headers->pdata[i];
-			encoded_value = g_mime_utils_8bit_header_encode (header->value);
-			
-			buf = g_mime_utils_header_printf ("%s: %s\n", header->name, encoded_value);
-			g_free (encoded_value);
-			
-			g_string_append (string, buf);
-			g_free (buf);
-		}
-	}
-	
-	/* create the standard headers */
-	if (!message->header->date)
-		g_mime_message_set_date (message, time (NULL), 0);
-	date = g_mime_message_get_date_string (message);
-	buf = g_mime_utils_header_printf ("Date: %s\n", date);
-	g_string_append (string, buf);
-	g_free (date);
-	g_free (buf);
-	
-	buf = g_mime_utils_header_printf ("From: %s\n", message->header->from ? message->header->from : "");
-	g_string_append (string, buf);
-	g_free (buf);
-	
-	if (message->header->reply_to) {
-		buf = g_mime_utils_header_printf ("Reply-To: %s\n", message->header->reply_to);
-		g_string_append (string, buf);
-		g_free (buf);
-	}
-	
-	recipients = g_mime_message_get_recipients (message, GMIME_RECIPIENT_TYPE_TO);
-	if (recipients) {
-		GString *recip;
-		
-		recip = g_string_new ("To: ");
-		while (TRUE) {
-			InternetAddress *ia;
-			gchar *address;
-			
-			ia = recipients->data;
-			address = internet_address_to_string (ia, TRUE);
-			g_string_append (recip, address);
-			g_free (address);
-			
-			recipients = recipients->next;
-			if (recipients)
-				g_string_append (recip, ", ");
-			else
-				break;
-		}
-		g_string_append (recip, "\n");
-		buf = g_mime_utils_header_fold (recip->str);
-		g_string_free (recip, TRUE);
-		
-		g_string_append (string, buf);
-		g_free (buf);
-	}
-	
-	recipients = g_mime_message_get_recipients (message, GMIME_RECIPIENT_TYPE_CC);
-	if (recipients) {
-		GString *recip;
-		
-		recip = g_string_new ("Cc: ");
-		while (TRUE) {
-			InternetAddress *ia;
-			gchar *address;
-			
-			ia = recipients->data;
-			address = internet_address_to_string (ia, TRUE);
-			g_string_append (recip, address);
-			g_free (address);
-			
-			recipients = recipients->next;
-			if (recipients)
-				g_string_append (recip, ", ");
-			else
-				break;
-		}
-		g_string_append (recip, "\n");
-		buf = g_mime_utils_header_fold (recip->str);
-		g_string_free (recip, TRUE);
-		
-		g_string_append (string, buf);
-		g_free (buf);
-	}
-	
-	subject = g_mime_utils_8bit_header_encode (message->header->subject);
-	buf = g_mime_utils_header_printf ("Subject: %s\n", subject ? subject : "");
-	g_string_append (string, buf);
-	g_free (subject);
-	g_free (buf);
-	
-	if (message->header->message_id) {
-		gchar *message_id;
-		
-		message_id = g_mime_utils_8bit_header_encode (message->header->message_id);
-		buf = g_mime_utils_header_printf ("Message-Id: %s\n", message_id ? message_id : "");
-		g_string_append (string, buf);
-		g_free (message_id);
-		g_free (buf);
-	}
-	
-	str = string->str;
-	g_string_free (string, FALSE);
-	
-	return str;
+	if (message->mime_part) {
+		g_mime_stream_write_string (stream, "MIME-Version: 1.0\n");
+		g_mime_part_write_to_stream (message->mime_part, stream);
+	} else
+		g_mime_stream_write (stream, "\n", 1);
 }
 
 
@@ -606,27 +580,20 @@ create_header (GMimeMessage *message)
 gchar *
 g_mime_message_to_string (GMimeMessage *message)
 {
-	GString *string;
-	gchar *str, *header, *body;
+	GMimeStream *stream;
+	GByteArray *array;
+	char *str;
 	
 	g_return_val_if_fail (message != NULL, NULL);
 	
-	header = create_header (message);
-	if (!header)
-		return NULL;
-	
-	string = g_string_new (header);
-	g_free (header);
-	
-	body = g_mime_part_to_string (message->mime_part, TRUE);
-	if (body)
-		g_string_append (string, body);
-	else
-		g_string_append (string, "\n");
-	g_free (body);
-	
-	str = string->str;
-	g_string_free (string, FALSE);
+	array = g_byte_array_new ();
+	stream = g_mime_stream_mem_new ();
+	g_mime_stream_mem_set_byte_array (GMIME_STREAM_MEM (stream), array);
+	g_mime_message_write_to_stream (message, stream);
+	g_mime_stream_unref (stream);
+	g_byte_array_append (array, "", 1);
+	str = array->data;
+	g_byte_array_free (array, FALSE);
 	
 	return str;
 }
@@ -714,6 +681,7 @@ multipart_get_body (GMimePart *multipart, gboolean want_plain, gboolean *is_html
 	return body;
 }
 
+
 /**
  * g_mime_message_get_body: Return the body of the message
  * @message: MIME Message
@@ -763,7 +731,9 @@ g_mime_message_get_body (const GMimeMessage *message, gboolean want_plain, gbool
 gchar *
 g_mime_message_get_headers (GMimeMessage *message)
 {
-	return create_header (message);
+	g_return_val_if_fail (message != NULL, NULL);
+	
+	return g_mime_header_to_string (message->header->headers);
 }
 
 
