@@ -21,9 +21,14 @@
  *
  */
 
-#include "gmime-utils.h"
-#include "gmime-part.h"
+#ifdef HAVE_CONFIG_H
 #include <config.h>
+#endif
+
+#include "gmime-utils.h"
+#include "gmime-table-private.h"
+#include "gmime-part.h"
+#include "gmime-charset.h"
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -31,10 +36,15 @@
 
 #define d(x)
 
+#define GMIME_UUENCODE_CHAR(c) ((c) ? (c) + ' ' : '`')
 #define	GMIME_UUDECODE_CHAR(c) (((c) - ' ') & 077)
 
 #define GMIME_FOLD_LEN  76
 
+#ifndef HAVE_ISBLANK
+#define isblank(c) (c == ' ' || c == '\t')
+#endif
+
 static char *base64_alphabet =
 "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
 
@@ -43,25 +53,6 @@ static unsigned char tohex[16] = {
 	'8', '9', 'A', 'B', 'C', 'D', 'E', 'F'
 };
 
-static unsigned short gmime_special_table[256] = {
-	  5,  5,  5,  5,  5,  5,  5,  5,  5,231,  7,  5,  5, 39,  5,  5,
-	  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,
-	242,448, 76,192,192,192,192,192, 76, 76,448,448, 76,448, 72,324,
-	448,448,448,448,448,448,448,448,448,448, 76, 76, 76,  4, 76, 68,
-	 76,448,448,448,448,448,448,448,448,448,448,448,448,448,448,448,
-	448,448,448,448,448,448,448,448,448,448,448,108,236,108,192, 64,
-	192,448,448,448,448,448,448,448,448,448,448,448,448,448,448,448,
-	448,448,448,448,448,448,448,448,448,448,448,192,192,192,192,  5,
-	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
-	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
-	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
-	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
-	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
-	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
-	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
-	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
-};
-
 static unsigned char gmime_base64_rank[256] = {
 	255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
 	255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
@@ -101,109 +92,6 @@ static unsigned char gmime_uu_rank[256] = {
 };
 
 
-enum {
-	IS_CTRL		= 1<<0,
-	IS_LWSP		= 1<<1,
-	IS_TSPECIAL	= 1<<2,
-	IS_SPECIAL	= 1<<3,
-	IS_SPACE	= 1<<4,
-	IS_DSPECIAL	= 1<<5,
-	IS_QPSAFE	= 1<<6,
-	IS_ESAFE	= 1<<7,	/* encoded word safe */
-	IS_PSAFE	= 1<<8,	/* encoded word in phrase safe */
-};
-
-#define is_ctrl(x) ((gmime_special_table[(unsigned char)(x)] & IS_CTRL) != 0)
-#define is_lwsp(x) ((gmime_special_table[(unsigned char)(x)] & IS_LWSP) != 0)
-#define is_tspecial(x) ((gmime_special_table[(unsigned char)(x)] & IS_TSPECIAL) != 0)
-#define is_type(x, t) ((gmime_special_table[(unsigned char)(x)] & (t)) != 0)
-#define is_ttoken(x) ((gmime_special_table[(unsigned char)(x)] & (IS_TSPECIAL|IS_LWSP|IS_CTRL)) == 0)
-#define is_atom(x) ((gmime_special_table[(unsigned char)(x)] & (IS_SPECIAL|IS_SPACE|IS_CTRL)) == 0)
-#define is_dtext(x) ((gmime_special_table[(unsigned char)(x)] & IS_DSPECIAL) == 0)
-#define is_fieldname(x) ((gmime_special_table[(unsigned char)(x)] & (IS_CTRL|IS_SPACE)) == 0)
-#define is_qpsafe(x) ((gmime_special_table[(unsigned char)(x)] & IS_QPSAFE) != 0)
-#define is_especial(x) ((gmime_special_table[(unsigned char)(x)] & IS_ESPECIAL) != 0)
-#define is_psafe(x) ((gmime_special_table[(unsigned char)(x)] & IS_PSAFE) != 0)
-
-#ifndef HAVE_ISBLANK
-#define isblank(c) ((c) == ' ' || (c) == '\t')
-#endif /* HAVE_ISBLANK */
-
-#define CHARS_LWSP " \t\n\r"               /* linear whitespace chars */
-#define CHARS_TSPECIAL "()<>@,;:\\\"/[]?="
-#define CHARS_SPECIAL "()<>@,;:\\\".[]"
-#define CHARS_CSPECIAL "()\\\r"	           /* not in comments */
-#define CHARS_DSPECIAL "[]\\\r \t"	   /* not in domains */
-#define CHARS_ESPECIAL "()<>@,;:\"/[]?.="  /* encoded word specials (rfc2047 5.1) */
-#define CHARS_PSPECIAL "!*+-/"             /* encoded phrase specials (rfc2047 5.3) */
-
-#ifdef BUILD_TABLE
-/* code to rebuild the gmime_special_table */
-static void
-header_remove_bits (gushort bit, guchar *vals)
-{
-	gint i;
-	
-	for (i = 0; vals[i]; i++)
-		gmime_special_table[vals[i]] &= ~bit;
-}
-
-static void
-header_init_bits (gushort bit, gushort bitcopy, gboolean remove, guchar *vals)
-{
-	gint i, len = strlen (vals);
-	
-	if (!remove) {
-		for (i = 0; i < len; i++) {
-			gmime_special_table[vals[i]] |= bit;
-		}
-		if (bitcopy) {
-			for (i = 0; i < 256; i++) {
-				if (gmime_special_table[i] & bitcopy)
-					gmime_special_table[i] |= bit;
-			}
-		}
-	} else {
-		for (i = 0; i < 256; i++)
-			gmime_special_table[i] |= bit;
-		for (i = 0; i < len; i++) {
-			gmime_special_table[vals[i]] &= ~bit;
-		}
-		if (bitcopy) {
-			for (i = 0; i < 256; i++) {
-				if (gmime_special_table[i] & bitcopy)
-					gmime_special_table[i] &= ~bit;
-			}
-		}
-	}
-}
-
-static void
-header_decode_init (void)
-{
-	gint i;
-	
-	for (i = 0; i < 256; i++) {
-		gmime_special_table[i] = 0;
-		if (i < 32)
-			gmime_special_table[i] |= IS_CTRL;
-		if ((i >= 33 && i <= 60) || (i >= 62 && i <= 126) || i == 32 || i == 9)
-			gmime_special_table[i] |= (IS_QPSAFE | IS_ESAFE);
-		if ((i >= '0' && i <= '9') || (i >= 'a' && i <= 'z') || (i >= 'A' && i <= 'Z'))
-			gmime_special_table[i] |= IS_PSAFE;
-	}
-	
-	gmime_special_table[127] |= IS_CTRL;
-	gmime_special_table[' '] |= IS_SPACE;
-	header_init_bits (IS_LWSP, 0, FALSE, CHARS_LWSP);
-	header_init_bits (IS_TSPECIAL, IS_CTRL, FALSE, CHARS_TSPECIAL);
-	header_init_bits (IS_SPECIAL, 0, FALSE, CHARS_SPECIAL);
-	header_init_bits (IS_DSPECIAL, 0, FALSE, CHARS_DSPECIAL);
-	header_remove_bits (IS_ESAFE, CHARS_ESPECIAL);
-	header_init_bits (IS_PSAFE, 0, FALSE, CHARS_PSPECIAL);
-}
-#endif /* BUILD_TABLE */
-
 /* hrm, is there a library for this shit? */
 static struct {
 	char *name;
@@ -273,10 +161,10 @@ datetok (const gchar *date)
 	start = (gchar *) date;
 	while (*start) {
 		/* kill leading whitespace */
-		for ( ; *start && isspace (*start); start++);
+		for ( ; *start && isspace ((int)*start); start++);
 		
 		/* find the end of this token */
-		for (end = start; *end && !isspace (*end); end++);
+		for (end = start; *end && !isspace ((int)*end); end++);
 		
 		token = g_strndup (start, (end - start));
 		
@@ -346,7 +234,7 @@ get_mday (gchar *str)
 	g_return_val_if_fail (str != NULL, -1);
 	
 	for (p = str; *p; p++)
-		if (!isdigit (*p))
+		if (!isdigit ((int)*p))
 			return -1;
 	
 	mday = atoi (str);
@@ -378,7 +266,7 @@ get_year (const gchar *str)
 	const gchar * p;
 	
 	for (p = str; *p; p++)
-		if (!isdigit (*p))
+		if (!isdigit ((int)*p))
 			return -1;
 	
 	year = atoi (str);
@@ -402,12 +290,11 @@ get_time (const gchar *in, gint *hour, gint *min, gint *sec)
 	for (p = in; *p && digits; p++) {
 		if (*p == ':')
 			colons++;
-		else if (!isdigit (*p))
+		else if (!isdigit ((int)*p))
 			digits = FALSE;
 	}
 	
-/* Ameol software doesn't put the seconds on the time */
-
+	/* Ameol software doesn't put the seconds on the time */
 	if (!digits || (colons != 2 && colons != 1))
 		return FALSE;
 	
@@ -655,6 +542,28 @@ g_mime_utils_header_printf (const gchar *format, ...)
 	return ret;
 }
 
+static gboolean
+need_quotes (const char *string)
+{
+	gboolean quoted = FALSE;
+	const char *inptr;
+	
+	inptr = string;
+	
+	while (*inptr) {
+		if (*inptr == '\\')
+			inptr++;
+		else if (*inptr == '"')
+			quoted = !quoted;
+		else if (!quoted && (is_tspecial (*inptr) || *inptr == '.'))
+			return TRUE;
+		
+		if (*inptr)
+			inptr++;
+	}
+	
+	return FALSE;
+}
 
 /**
  * g_mime_utils_quote_string: Quote a string.
@@ -668,18 +577,16 @@ g_mime_utils_header_printf (const gchar *format, ...)
 gchar *
 g_mime_utils_quote_string (const gchar *string)
 {
-	GString *out;
+	gboolean quote;
+	const gchar *c;
 	gchar *qstring;
-	guchar *c;
-	gboolean quote = FALSE;
+	GString *out;
 	
 	out = g_string_new ("");
+	quote = need_quotes (string);
 	
-	for (c = (guchar *) string; *c; c++) {
-		if (is_tspecial (*c))
-			quote = TRUE;
-		
-		if (*c == '"' || *c == '\\')
+	for (c = string; *c; c++) {
+		if ((*c == '"' && quote) || *c == '\\')
 			g_string_append_c (out, '\\');
 		
 		g_string_append_c (out, *c);
@@ -833,16 +740,23 @@ quoted_decode (const guchar *in, gint len, guchar *out)
 }
 
 static gboolean
-is_8bit_word_encoded (const guchar * word, guint len)
+is_8bit_word_encoded (const guchar *atom)
 {
-	return len>=7 && word[0]=='=' && word[1]=='?' && word[len-2]=='?' && word[len-1]=='=';
+	guint len = strlen (atom);
+	
+	return len >= 7 && !strncmp (atom, "=?", 2) && !strncmp (atom + len - 2, "?=", 2);
 }
 
 static guchar *
-decode_encoded_8bit_word (const guchar *word, guint len)
+decode_encoded_8bit_word (const guchar *word)
 {
-	guchar * inptr = (guchar *) word + 2;
-	guchar * inend = (guchar *) word + len - 2;
+	guchar *inptr, *inend;
+	guint len;
+	
+	len = strlen (word);
+	
+	inptr = (guchar *) word + 2;
+	inend = (guchar *) word + len - 2;
 	
 	inptr = memchr (inptr, '?', inend - inptr);
 	if (inptr && inptr[2] == '?') {
@@ -894,95 +808,104 @@ decode_encoded_8bit_word (const guchar *word, guint len)
 gchar *
 g_mime_utils_8bit_header_decode (const guchar *in)
 {
-	GString *out, *whtspc, *word;
+	GString *out, *lwsp, *atom;
 	const guchar *inptr;
 	guchar *decoded;
 	gboolean last_was_encoded = FALSE;
 	gboolean last_was_space = FALSE;
-
-	/* primitive encoding check to avoid doing more work */
-	if (in!=NULL && strstr(in,"=?")==NULL)
-		return g_strdup (in);
-
+	
 	out = g_string_sized_new (256);
-	whtspc = g_string_sized_new (256);
-	word = g_string_sized_new (256);
+	lwsp = g_string_sized_new (256);
+	atom = g_string_sized_new (256);
 	inptr = in;
 	
 	while (inptr && *inptr) {
 		guchar c = *inptr++;
 		
-		if (isspace (c) && !last_was_space) {
-			/* we reached the end of a 'word' */
+		if (!is_atom (c) && !last_was_space) {
+			/* we reached the end of an atom */
 			gboolean was_encoded;
-			const guchar *cword;
 			guchar *dword = NULL;
-
-			if ((was_encoded = is_8bit_word_encoded (word->str, word->len)))
-				cword = dword = decode_encoded_8bit_word (word->str, word->len);
+			const guchar *word;
+			
+			if ((was_encoded = is_8bit_word_encoded (atom->str)))
+				word = dword = decode_encoded_8bit_word (atom->str);
 			else
-				cword = word->str;
-
-			if (cword!=NULL) {
+				word = atom->str;
+			
+			if (word) {
 				if (!(last_was_encoded && was_encoded)) {
 					/* rfc2047 states that you
                                            must ignore all whitespace
                                            between encoded words */
-					g_string_append (out, whtspc->str);
+					g_string_append (out, lwsp->str);
 				}
 				
-				g_string_append (out, cword);
+				g_string_append (out, word);
 				g_free (dword);
 			} else {
 				was_encoded = FALSE;
-				g_string_append (out, whtspc->str);
-				g_string_append (out, word->str);
+				g_string_append (out, lwsp->str);
+				g_string_append (out, atom->str);
 			}
 			
 			last_was_encoded = was_encoded;
 			
-			g_string_truncate (whtspc, 0);
-			g_string_truncate (word, 0);
+			g_string_truncate (lwsp, 0);
+			g_string_truncate (atom, 0);
+			
+			if (is_lwsp (c)) {
+				g_string_append_c (lwsp, c);
+				last_was_space = TRUE;
+			} else {
+				/* This is mostly here for interoperability with broken
+                                   mailers that might do something stupid like:
+                                   =?iso-8859-1?Q?blah?=:\t=?iso-8859-1?Q?I_am_broken?= */
+				g_string_append_c (out, c);
+				last_was_encoded = FALSE;
+				last_was_space = FALSE;
+			}
+			
+			continue;
 		}
 		
-		if (isspace (c)) {
-			g_string_append_c (whtspc, c);
-			last_was_space = TRUE;
-		} else {
-			g_string_append_c (word, c);
+		if (is_atom (c)) {
+			g_string_append_c (atom, c);
 			last_was_space = FALSE;
+		} else {
+			g_string_append_c (lwsp, c);
+			last_was_space = TRUE;
 		}
 	}
 	
-	if (word->len || whtspc->len) {
+	if (atom->len || lwsp->len) {
 		gboolean was_encoded;
-		const guchar *cword;
 		guchar *dword = NULL;
-
-		if ((was_encoded = is_8bit_word_encoded (word->str, word->len)))
-			cword = dword = decode_encoded_8bit_word (word->str, word->len);
-		else
-			cword = word->str;
+		const guchar *word;
 		
-		if (cword != NULL) {
+		if ((was_encoded = is_8bit_word_encoded (atom->str)))
+			word = dword = decode_encoded_8bit_word (atom->str);
+		else
+			word = atom->str;
+		
+		if (word) {
 			if (!(last_was_encoded && was_encoded)) {
 				/* rfc2047 states that you
 				   must ignore all whitespace
 				   between encoded words */
-				g_string_append (out, whtspc->str);
+				g_string_append (out, lwsp->str);
 			}
 			
-			g_string_append (out, cword);
+			g_string_append (out, word);
 			g_free (dword);
 		} else {
-			g_string_append (out, whtspc->str);
-			g_string_append (out, word->str);
+			g_string_append (out, lwsp->str);
+			g_string_append (out, atom->str);
 		}
-
 	}
 	
-	g_string_free (whtspc, TRUE);
-	g_string_free (word, TRUE);
+	g_string_free (lwsp, TRUE);
+	g_string_free (atom, TRUE);
 	
 	decoded = out->str;
 	g_string_free (out, FALSE);
@@ -1004,9 +927,9 @@ quoted_encode (const guchar *in, gint len, guchar *out, gushort safemask)
 	
 	while (inptr < inend) {
 		c = *inptr++;
-		if (gmime_special_table[c] & safemask && c != '\t') {
-			if (c == ' ')
-				c = '_';
+		if (c == ' ') {
+			*outptr++ = '_';
+		} else if (gmime_special_table[c] & safemask) {
 			*outptr++ = c;
 		} else {
 			*outptr++ = '=';
@@ -1065,7 +988,7 @@ encode_8bit_word (const guchar *word, gushort safemask, gboolean *this_was_encod
 	if (this_was_encoded)
 		*this_was_encoded = TRUE;
 	
-	return g_strdup_printf ("=?iso-8859-1?%c?%s?=", encoding, encoded);
+	return g_strdup_printf ("=?%s?%c?%s?=", g_mime_charset_locale_name (), encoding, encoded);	
 }
 
 
@@ -1093,7 +1016,7 @@ g_mime_utils_8bit_header_encode_phrase (const guchar *in)
 gchar *
 g_mime_utils_8bit_header_encode (const guchar *in)
 {
-	GString *out, *word, *whtspc;
+	GString *out, *word, *lwsp;
 	guchar *inptr;
 	guchar *encoded;
 	gboolean is8bit = FALSE;
@@ -1102,7 +1025,7 @@ g_mime_utils_8bit_header_encode (const guchar *in)
 	
 	out = g_string_new ("");
 	word = g_string_new ("");
-	whtspc = g_string_new ("");
+	lwsp = g_string_new ("");
 	inptr = (guchar *) in;
 	
 	while (inptr && *inptr) {
@@ -1120,36 +1043,31 @@ g_mime_utils_8bit_header_encode (const guchar *in)
 			/* append any whitespace */
 			if (last_was_encoded && this_was_encoded) {
 				/* we need to encode the whitespace */
-				guchar *ewhtspc;
+				guchar *elwsp;
 				gint len;
 				
-				ewhtspc = alloca (whtspc->len * 3 + 4);
-				len = quoted_encode (whtspc->str, whtspc->len, ewhtspc, IS_ESAFE);
-				ewhtspc[len] = '\0';
+				elwsp = alloca (lwsp->len * 3 + 4);
+				len = quoted_encode (lwsp->str, lwsp->len, elwsp, IS_SPACE);
+				elwsp[len] = '\0';
 				
-				g_string_append (out, " =?iso-8859-1?q?");
-				g_string_append (out, ewhtspc);
-				g_string_append (out, "?= ");
+				g_string_sprintfa (out, " =?%s?q?%s?= ", g_mime_charset_locale_name (), elwsp);
 			} else {
-				g_string_append (out, whtspc->str);
+				g_string_append (out, lwsp->str);
 			}
 			
 			/* append the encoded word */
 			g_string_append (out, eword);
 			g_free (eword);
 			
-			g_string_free (whtspc, TRUE);
-			whtspc = g_string_new ("");
-			
-			g_string_free (word, TRUE);
-			word = g_string_new ("");
+			g_string_truncate (lwsp, 0);
+			g_string_truncate (word, 0);
 			
 			last_was_encoded = this_was_encoded;
 			is8bit = FALSE;
 		}
 		
 		if (isspace (c)) {
-			g_string_append_c (whtspc, c);
+			g_string_append_c (lwsp, c);
 			last_was_space = TRUE;
 		} else {
 			if (c > 127)
@@ -1160,7 +1078,7 @@ g_mime_utils_8bit_header_encode (const guchar *in)
 		}
 	}
 	
-	if (word->len || whtspc->len) {
+	if (word->len || lwsp->len) {
 		gboolean this_was_encoded = FALSE;
 		guchar *eword;
 		
@@ -1172,18 +1090,16 @@ g_mime_utils_8bit_header_encode (const guchar *in)
 		/* append any whitespace */
 		if (last_was_encoded && this_was_encoded) {
 			/* we need to encode the whitespace */
-			guchar *ewhtspc;
+			guchar *elwsp;
 			gint len;
 			
-			ewhtspc = alloca (whtspc->len * 3 + 4);
-			len = quoted_encode (whtspc->str, whtspc->len, ewhtspc, IS_ESAFE);
-			ewhtspc[len] = '\0';
+			elwsp = alloca (lwsp->len * 3 + 4);
+			len = quoted_encode (lwsp->str, lwsp->len, elwsp, IS_SPACE);
+			elwsp[len] = '\0';
 			
-			g_string_append (out, " =?iso-8859-1?q?");
-			g_string_append (out, ewhtspc);
-			g_string_append (out, "?= ");
+			g_string_sprintfa (out, " =?%s?q?%s?= ", g_mime_charset_locale_name (), elwsp);
 		} else {
-			g_string_append (out, whtspc->str);
+			g_string_append (out, lwsp->str);
 		}
 		
 		/* append the encoded word */
@@ -1191,7 +1107,7 @@ g_mime_utils_8bit_header_encode (const guchar *in)
 		g_free (eword);
 	}
 	
-	g_string_free (whtspc, TRUE);
+	g_string_free (lwsp, TRUE);
 	g_string_free (word, TRUE);
 	
 	encoded = out->str;
@@ -1214,18 +1130,18 @@ g_mime_utils_8bit_header_encode (const guchar *in)
  * bit.
  **/
 gint
-g_mime_utils_base64_encode_close (const guchar *in, gint inlen, guchar *out, gint *state, gint *save)
+g_mime_utils_base64_encode_close (const guchar *in, gint inlen, guchar *out, gint *state, guint32 *save)
 {
-	guchar *outptr = out;
-	gint c1, c2;
+	unsigned char *outptr = out;
+	int c1, c2;
 	
 	if (inlen > 0)
 		outptr += g_mime_utils_base64_encode_step (in, inlen, outptr, state, save);
 	
-	c1 = ((guchar *)save)[1];
-	c2 = ((guchar *)save)[2];
+	c1 = ((unsigned char *)save)[1];
+	c2 = ((unsigned char *)save)[2];
 	
-	switch (((guchar *)save)[0]) {
+	switch (((unsigned char *)save)[0]) {
 	case 2:
 		outptr[2] = base64_alphabet [(c2 & 0x0f) << 2];
 		goto skip;
@@ -1261,10 +1177,10 @@ g_mime_utils_base64_encode_close (const guchar *in, gint inlen, guchar *out, gin
  * invocation).
  **/
 gint
-g_mime_utils_base64_encode_step (const guchar *in, gint inlen, guchar *out, gint *state, gint *save)
+g_mime_utils_base64_encode_step (const guchar *in, gint inlen, guchar *out, gint *state, guint32 *save)
 {
-	const register guchar *inptr;
-	register guchar *outptr;
+	const register unsigned char *inptr;
+	register unsigned char *outptr;
 	
 	if (inlen <= 0)
 		return 0;
@@ -1274,17 +1190,17 @@ g_mime_utils_base64_encode_step (const guchar *in, gint inlen, guchar *out, gint
 	
 	d(printf("we have %d chars, and %d saved chars\n", inlen, ((gchar *)save)[0]));
 	
-	if (inlen + ((guchar *)save)[0] > 2) {
-		const guchar *inend = in + inlen - 2;
-		register gint c1=0, c2=0, c3=0;
-		register gint already;
+	if (inlen + ((unsigned char *)save)[0] > 2) {
+		const unsigned char *inend = in + inlen - 2;
+		register int c1 = 0, c2 = 0, c3 = 0;
+		register int already;
 		
 		already = *state;
 		
 		switch (((gchar *)save)[0]) {
-		case 1:	c1 = ((guchar *)save)[1]; goto skip1;
-		case 2:	c1 = ((guchar *)save)[1];
-			c2 = ((guchar *)save)[2]; goto skip2;
+		case 1:	c1 = ((unsigned char *)save)[1]; goto skip1;
+		case 2:	c1 = ((unsigned char *)save)[1];
+			c2 = ((unsigned char *)save)[2]; goto skip2;
 		}
 		
 		/* yes, we jump into the loop, no i'm not going to change it, its beautiful! */
@@ -1296,7 +1212,7 @@ g_mime_utils_base64_encode_step (const guchar *in, gint inlen, guchar *out, gint
 			c3 = *inptr++;
 			*outptr++ = base64_alphabet [c1 >> 2];
 			*outptr++ = base64_alphabet [(c2 >> 4) | ((c1 & 0x3) << 4)];
-			*outptr++ = base64_alphabet [((c2 &0x0f) << 2) | (c3 >> 6)];
+			*outptr++ = base64_alphabet [((c2 & 0x0f) << 2) | (c3 >> 6)];
 			*outptr++ = base64_alphabet [c3 & 0x3f];
 			/* this is a bit ugly ... */
 			if ((++already) >= 19) {
@@ -1305,7 +1221,7 @@ g_mime_utils_base64_encode_step (const guchar *in, gint inlen, guchar *out, gint
 			}
 		}
 		
-		((guchar *)save)[0] = 0;
+		((unsigned char *)save)[0] = 0;
 		inlen = 2 - (inptr - inend);
 		*state = already;
 	}
@@ -1345,37 +1261,37 @@ g_mime_utils_base64_encode_step (const guchar *in, gint inlen, guchar *out, gint
  * Returns the number of bytes decoded (which have been dumped in #out).
  **/
 gint
-g_mime_utils_base64_decode_step (const guchar *in, gint inlen, guchar *out, gint *state, guint *save)
+g_mime_utils_base64_decode_step (const guchar *in, gint inlen, guchar *out, gint *state, guint32 *save)
 {
-	const register guchar *inptr;
-	register guchar *outptr;
-	const guchar *inend;
-	guchar c;
-	register guint v;
+	const register unsigned char *inptr;
+	register unsigned char *outptr;
+	const unsigned char *inend;
+	register guint32 saved;
+	unsigned char c;
 	int i;
 	
 	inend = in + inlen;
 	outptr = out;
 	
 	/* convert 4 base64 bytes to 3 normal bytes */
-	v = *save;
+	saved = *save;
 	i = *state;
 	inptr = in;
 	while (inptr < inend) {
 		c = gmime_base64_rank[*inptr++];
 		if (c != 0xff) {
-			v = (v << 6) | c;
+			saved = (saved << 6) | c;
 			i++;
 			if (i == 4) {
-				*outptr++ = v >> 16;
-				*outptr++ = v >> 8;
-				*outptr++ = v;
+				*outptr++ = saved >> 16;
+				*outptr++ = saved >> 8;
+				*outptr++ = saved;
 				i = 0;
 			}
 		}
 	}
 	
-	*save = v;
+	*save = saved;
 	*state = i;
 	
 	/* quick scan back for '=' on the end somewhere */
@@ -1395,6 +1311,159 @@ g_mime_utils_base64_decode_step (const guchar *in, gint inlen, guchar *out, gint
 }
 
 
+/**
+ * g_mime_utils_uuencode_close: uuencode a chunk of data
+ * @in: input stream
+ * @inlen: input stream length
+ * @out: output stream
+ * @uubuf: temporary buffer of 60 bytes
+ * @state: holds the number of bits that are stored in @save
+ * @save: leftover bits that have not yet been encoded
+ * @uulen: holds the value of the length-char which is used to calculate
+ *         how many more chars need to be decoded for that 'line'
+ *
+ * Returns the number of bytes encoded. Call this when finished
+ * encoding data with uuencode_step to flush off the last little
+ * bit.
+ **/
+gint
+g_mime_utils_uuencode_close (const guchar *in, gint inlen, guchar *out, guchar *uubuf, gint *state, guint32 *save, gchar *uulen)
+{
+	register unsigned char *outptr, *bufptr;
+	register guint32 saved;
+	int i;
+	
+	outptr = out;
+	
+	if (inlen > 0)
+		outptr += g_mime_utils_uuencode_step (in, inlen, out, uubuf, state, save, uulen);
+	
+	bufptr = uubuf + ((*uulen / 3) * 4);
+	saved = *save;
+	i = *state;
+	
+	if (i > 0) {
+		while (i < 3) {
+			saved <<= 8 | 0;
+			i++;
+		}
+		
+		if (i == 3) {
+			/* convert 3 normal bytes into 4 uuencoded bytes */
+			unsigned char b0, b1, b2;
+			
+			b0 = saved >> 16;
+			b1 = saved >> 8 & 0xff;
+			b2 = saved & 0xff;
+			
+			*bufptr++ = GMIME_UUENCODE_CHAR ((b0 >> 2) & 0x3f);
+			*bufptr++ = GMIME_UUENCODE_CHAR (((b0 << 4) | ((b1 >> 4) & 0xf)) & 0x3f);
+			*bufptr++ = GMIME_UUENCODE_CHAR (((b1 << 2) | ((b2 >> 6) & 0x3)) & 0x3f);
+			*bufptr++ = GMIME_UUENCODE_CHAR (b2 & 0x3f);
+		}
+	}
+	
+	if (*uulen || *state) {
+		int cplen = (((*uulen + (*state ? 3 : 0)) / 3) * 4);
+		
+		*outptr++ = GMIME_UUENCODE_CHAR (*uulen + *state);
+		memcpy (outptr, uubuf, cplen);
+		outptr += cplen;
+		*outptr++ = '\n';
+		*uulen = 0;
+	}
+	
+	*outptr++ = GMIME_UUENCODE_CHAR (*uulen);
+	*outptr++ = '\n';
+	
+	*save = 0;
+	*state = 0;
+	
+	return (outptr - out);
+}
+
+
+/**
+ * g_mime_utils_uuencode_step: uuencode a chunk of data
+ * @in: input stream
+ * @inlen: input stream length
+ * @out: output stream
+ * @uubuf: temporary buffer of 60 bytes
+ * @state: holds the number of bits that are stored in @save
+ * @save: leftover bits that have not yet been encoded
+ * @uulen: holds the value of the length-char which is used to calculate
+ *         how many more chars need to be decoded for that 'line'
+ *
+ * Returns the number of bytes encoded. Performs an 'encode step',
+ * only encodes blocks of 45 characters to the output at a time, saves
+ * left-over state in @uubuf, @state and @save (initialize to 0 on first
+ * invocation).
+ **/
+gint
+g_mime_utils_uuencode_step (const guchar *in, gint inlen, guchar *out, guchar *uubuf, gint *state, guint32 *save, gchar *uulen)
+{
+	register unsigned char *outptr, *bufptr;
+	const register unsigned char *inptr;
+	const unsigned char *inend;
+	register guint32 saved;
+	int i;
+	
+	if (*uulen <= 0)
+		*uulen = 0;
+	
+	inptr = in;
+	inend = in + inlen;
+	
+	outptr = out;
+	
+	bufptr = uubuf + ((*uulen / 3) * 4);
+	
+	saved = *save;
+	i = *state;
+	
+	while (inptr < inend) {
+		while (*uulen < 45 && inptr < inend) {
+			while (i < 3 && inptr < inend) {
+				saved = (saved << 8) | *inptr++;
+				i++;
+			}
+			
+			if (i == 3) {
+				/* convert 3 normal bytes into 4 uuencoded bytes */
+				unsigned char b0, b1, b2;
+				
+				b0 = saved >> 16;
+				b1 = saved >> 8 & 0xff;
+				b2 = saved & 0xff;
+				
+				*bufptr++ = GMIME_UUENCODE_CHAR ((b0 >> 2) & 0x3f);
+				*bufptr++ = GMIME_UUENCODE_CHAR (((b0 << 4) | ((b1 >> 4) & 0xf)) & 0x3f);
+				*bufptr++ = GMIME_UUENCODE_CHAR (((b1 << 2) | ((b2 >> 6) & 0x3)) & 0x3f);
+				*bufptr++ = GMIME_UUENCODE_CHAR (b2 & 0x3f);
+				
+				i = 0;
+				saved = 0;
+				*uulen += 3;
+			}
+		}
+		
+		if (*uulen >= 45) {
+			*outptr++ = GMIME_UUENCODE_CHAR (*uulen);
+			memcpy (outptr, uubuf, ((*uulen / 3) * 4));
+			outptr += ((*uulen / 3) * 4);
+			*outptr++ = '\n';
+			*uulen = 0;
+			bufptr = uubuf;
+		}
+	}
+	
+	*save = saved;
+	*state = i;
+	
+	return (outptr - out);
+}
+
+
 /**
  * g_mime_utils_uudecode_step: uudecode a chunk of data
  * @in: input stream
@@ -1405,17 +1474,17 @@ g_mime_utils_base64_decode_step (const guchar *in, gint inlen, guchar *out, gint
  * @uulen: holds the value of the length-char which is used to calculate
  *         how many more chars need to be decoded for that 'line'
  *
- * Returns the number of bytes decoded. Performs a 'uudecode step' on
+ * Returns the number of bytes decoded. Performs a 'decode step' on
  * a chunk of uuencoded data. Assumes the "begin <mode> <file name>"
  * line has been stripped off.
  **/
 gint
 g_mime_utils_uudecode_step (const guchar *in, gint inlen, guchar *out, gint *state, guint32 *save, gchar *uulen)
 {
-	const register guchar *inptr;
-	register guchar *outptr;
-	const guchar *inend;
-	guchar ch;
+	const register unsigned char *inptr;
+	register unsigned char *outptr;
+	const unsigned char *inend;
+	unsigned char ch;
 	register guint32 saved;
 	gboolean last_was_eoln;
 	int i;
@@ -1451,7 +1520,7 @@ g_mime_utils_uudecode_step (const guchar *in, gint inlen, guchar *out, gint *sta
 			i++;
 			if (i == 4) {
 				/* convert 4 uuencoded bytes to 3 normal bytes */
-				guchar b0, b1, b2, b3;
+				unsigned char b0, b1, b2, b3;
 				
 				b0 = saved >> 24;
 				b1 = saved >> 16 & 0xff;
@@ -1502,7 +1571,7 @@ g_mime_utils_uudecode_step (const guchar *in, gint inlen, guchar *out, gint *sta
 gint
 g_mime_utils_quoted_encode_close (const guchar *in, gint inlen, guchar *out, gint *state, gint *save)
 {
-	register guchar *outptr = out;
+	register unsigned char *outptr = out;
 	int last;
 	
 	if (inlen > 0)
@@ -1545,9 +1614,9 @@ g_mime_utils_quoted_encode_close (const guchar *in, gint inlen, guchar *out, gin
 gint
 g_mime_utils_quoted_encode_step (const guchar *in, gint inlen, guchar *out, gint *state, gint *save)
 {
-	const register guchar *inptr, *inend;
-	register guchar *outptr;
-	guchar c;
+	const register unsigned char *inptr, *inend;
+	register unsigned char *outptr;
+	unsigned char c;
 	register int sofar = *save;  /* keeps track of how many chars on a line */
 	register int last = *state;  /* keeps track if last char to end was a space cr etc */
 	
@@ -1644,10 +1713,10 @@ g_mime_utils_quoted_decode_step (const guchar *in, gint inlen, guchar *out, gint
 	 * Note: Trailing rubbish (at the end of input), like = or =x
 	 * or =\r will be lost.
 	 */
-	const register guchar *inptr;
-	register guchar *outptr;
-	const guchar *inend;
-	guchar c;
+	const register unsigned char *inptr;
+	register unsigned char *outptr;
+	const unsigned char *inend;
+	unsigned char c;
 	int state, save;
 	
 	inend = in + inlen;
