@@ -23,7 +23,7 @@
 #ifndef __NEWSRC_UI_H__
 #define __NEWSRC_UI_H__
 
-#include <pan/server.h>
+#include <pan/base/server.h>
 
 
 void  newsrc_import_dialog (Server*);
