@@ -36,18 +36,20 @@
 #include <sys/stat.h>
 #include <sys/time.h>
 #include <errno.h>
+#include <unistd.h>
 
 #include <glib.h>
 
-#include <pan/debug.h>
+#include <pan/base/debug.h>
+#include <pan/base/log.h>
+#include <pan/base/pan-glib-extensions.h>
+#include <pan/base/pan-i18n.h>
+#include <pan/base/server.h>
+
 #include <pan/gui.h>
-#include <pan/log.h>
-#include <pan/pan-glib-extensions.h>
-#include <pan/pan-i18n.h>
-#include <pan/nntp.h>
 #include <pan/prefs.h>
+#include <pan/nntp.h>
 #include <pan/queue.h>
-#include <pan/server.h>
 #include <pan/server-ui.h>
 
 /*********************
@@ -163,8 +165,10 @@ static GSList* todo_list = NULL;
 static GSList* queue_list = NULL;
 
 static pthread_mutex_t sock_lock;
-static pthread_mutex_t queue_lock;
+static pthread_mutex_t cond_mutex;
+static pthread_mutex_t todo_mutex;
 static pthread_cond_t qcond = PTHREAD_COND_INITIALIZER;
+static gboolean work_to_do = FALSE;
 
 static pthread_t queue_thread = 0;
 
@@ -192,19 +196,19 @@ void
 queue_wakeup (void)
 {
 	debug_enter ("queue_wakeup");
-        pthread_mutex_lock (&queue_lock);
-        pthread_cond_broadcast (&qcond);
-        pthread_mutex_unlock (&queue_lock);
+        pthread_mutex_lock (&cond_mutex);
+	work_to_do = TRUE;
+        pthread_cond_signal (&qcond);
+        pthread_mutex_unlock (&cond_mutex);
 	debug_exit ("queue_wakeup");
 }
 
-static gint
+static void
 server_online_status_changed_cb (gpointer obj, gpointer arg, gpointer data)
 {
 	debug_enter ("server_online_status_changed_cb");
 	queue_wakeup ();
 	debug_exit ("server_online_status_changed_cb");
-	return 0;
 }
 
 void
@@ -220,7 +224,8 @@ queue_init (int socket_qty, int tries)
 	queue_task_status_changed = pan_callback_new ();
 	queue_max_tries_changed = pan_callback_new ();
 
-	pthread_mutex_init (&queue_lock, NULL);
+	pthread_mutex_init (&cond_mutex, NULL);
+	pthread_mutex_init (&todo_mutex, NULL);
 	pthread_mutex_init (&sock_lock, NULL);
 
 	pan_callback_add (server_get_online_status_changed_callback(),
@@ -311,10 +316,20 @@ queue_add (Task * task)
 	queue_task_insert (task, task->high_priority ? 0 : -1);
 }
 
-const GSList*
+GPtrArray*
 queue_get_tasks (void)
 {
-	return queue_list;
+	GSList * l;
+	GPtrArray * a = g_ptr_array_new ();
+
+        pthread_mutex_lock (&todo_mutex);
+	for (l=queue_list; l!=NULL; l=l->next) {
+		pan_object_ref (PAN_OBJECT(l->data));
+		g_ptr_array_add (a, l->data);
+	}
+        pthread_mutex_unlock (&todo_mutex);
+
+	return a;
 }
 
 /**
@@ -448,13 +463,12 @@ real_queue_task_remove (Task* task)
 	debug1 (DEBUG_QUEUE, "queue_task_remove %p", task);
 
 	g_return_if_fail (task!=NULL);
-
-	if (task->thread_id != 0) /* if it's running, send an abort message */
+	if (task->is_running)
 	{
 		task_hint_abort (task);
 		queue_set_task_status (task, QUEUE_TASK_STATUS_ABORTING);
 	}
-	else /* not running; simple cleanup. queue_run_thread() zeroes out thread_id */
+	else /* not running; simple cleanup. queue_run_thread() zeroes out is_running */
 	{
 		const int old_index = g_slist_index (queue_list, task);
 		const gboolean had_socket = task->sock != NULL;
@@ -589,13 +603,10 @@ queue_set_task_status (Task* task, QueueTaskStatus status)
 	g_return_if_fail (task!=NULL);
 	g_return_if_fail (g_slist_index(queue_list, task) != -1);
 
-	g_hash_table_insert (
-		task_to_status, task,
-		GINT_TO_POINTER(status));
-
-	pan_callback_call (
-		queue_task_status_changed, task,
-		GINT_TO_POINTER(status)); 
+	g_hash_table_insert (task_to_status, task,
+	                     GINT_TO_POINTER(status));
+	pan_callback_call (queue_task_status_changed, task,
+	                   GINT_TO_POINTER(status)); 
 }
 
 /************
@@ -611,11 +622,11 @@ queue_new_todo (QueueTodoAction action, Task* task, int i1, int i2)
 	todo->index_2 = i2;
 	todo->action = action;
 
-	pthread_mutex_lock (&queue_lock);
+	pthread_mutex_lock (&todo_mutex);
 	todo_list = g_slist_append (todo_list, todo);
 	debug3 (DEBUG_QUEUE, "todo queue now has %d tasks (new task type: %d, int 1: %d)",
 		(int)g_slist_length(todo_list), action, i1);
-	pthread_mutex_unlock (&queue_lock);
+	pthread_mutex_unlock (&todo_mutex);
 
 	queue_wakeup ();
 }
@@ -666,9 +677,9 @@ queue_run_thread (void* vp)
 	if (aborted)
 		task->sock->error = TRUE;
 	socket_cleanup (task);
-
 	/* clean up the task */
-	task->thread_id = 0;
+	task->is_running = FALSE;
+
 	if (result==TASK_SUCCESS) /* success */
 	{
 		queue_task_remove (task);
@@ -677,6 +688,7 @@ queue_run_thread (void* vp)
 	{
 		++task->tries;
 		queue_set_task_status (task, QUEUE_TASK_STATUS_QUEUED);
+		queue_wakeup ();
 	}
 	else if (aborted) /* aborted */
 	{
@@ -698,13 +710,15 @@ queue_run_thread (void* vp)
 static void
 queue_run (Task* task)
 {
+	pthread_t thread;
 	g_return_if_fail (task!=NULL);
 	debug2 (DEBUG_QUEUE, "starting a thread to run task %p, sock %p", task, task->sock);
 
 	queue_set_task_status (task, QUEUE_TASK_STATUS_RUNNING);
 
-	pthread_create (&task->thread_id, NULL, (void*)queue_run_thread, task);
-	pthread_detach (task->thread_id);
+	task->is_running = TRUE;
+	pthread_create (&thread, NULL, (void*)queue_run_thread, task);
+	pthread_detach (thread);
 }
 
 
@@ -998,6 +1012,10 @@ socket_cleanup (Task* task)
 		else
 		{
 			/* otherwise throw it away */
+			pan_socket_putline (socket, "ccc\r\n");
+			nntp_disconnect (NULL, socket);
+			close (socket->sockfd);
+			socket->sockfd = 0;
 			pan_object_unref (PAN_OBJECT(socket));
 			gui_inc_connection_size (-1);
 			sockets = g_slist_remove (sockets, info);
@@ -1136,10 +1154,10 @@ queue_do_todo (void)
 	GSList * list = NULL;
 	GSList * l = NULL;
 
-	pthread_mutex_lock (&queue_lock);
+	pthread_mutex_lock (&todo_mutex);
 	list = todo_list;
 	todo_list = NULL;
-	pthread_mutex_unlock (&queue_lock);
+	pthread_mutex_unlock (&todo_mutex);
 
 	for (l=list; l!=NULL; l=l->next)
 	{
@@ -1192,28 +1210,30 @@ queue_do_todo (void)
 static void
 queue_mainloop (void)
 {
-	const int TIMEOUT_SECS = 20;
+	const int TIMEOUT_SECS = 15;
 
 	for (;;)
-        {
-		int retcode = 0;
+	{
+		gint retcode;
 		struct timeval now;
 		struct timespec timeout;
-		pthread_mutex_lock (&queue_lock);
+
+		pthread_mutex_lock (&cond_mutex);
 		gettimeofday (&now, NULL);
 		timeout.tv_sec = now.tv_sec + TIMEOUT_SECS;
-		timeout.tv_nsec = timeout.tv_sec;
-		retcode = pthread_cond_timedwait(&qcond, &queue_lock, &timeout);
-                pthread_mutex_unlock (&queue_lock);
+		timeout.tv_nsec = 0;
+		retcode = 0;
+		while (!work_to_do && retcode!=ETIMEDOUT) {
+			retcode = pthread_cond_timedwait(&qcond, &cond_mutex, &timeout);
+		}
+		work_to_do = FALSE;
+                pthread_mutex_unlock (&cond_mutex);
 
-		/* upkeep */
-		queue_do_todo ();
-
-                /* process any queue items */
-                queue_run_what_we_can ();
-
-                /* if timeout, do socket upkeep */
-                if (retcode==ETIMEDOUT)
+		if (retcode == ETIMEDOUT) {
 			sockets_upkeep ();
+		} else {
+			queue_do_todo ();
+                	queue_run_what_we_can ();
+		}
 	}
 }
