@@ -33,16 +33,18 @@
 
 #include <gmime/gmime-utils.h>
 
-#include <pan/article.h>
-#include <pan/gnksa.h>
-#include <pan/pan-glib-extensions.h>
-#include <pan/pan-i18n.h>
-#include <pan/status-item.h>
+#include <pan/base/article.h>
+#include <pan/base/gnksa.h>
+#include <pan/base/pan-glib-extensions.h>
+#include <pan/base/pan-i18n.h>
+#include <pan/base/status-item.h>
+
+#include <pan/identities/identity.h>
+#include <pan/identities/identity-manager.h>
 
 #include <pan/prefs.h>
 #include <pan/smtp.h>
 #include <pan/sockets.h>
-#include <pan/util.h>
 
 extern const gchar* sockread_err_msg;
 extern const gchar* sockwrite_err_msg;
@@ -180,7 +182,7 @@ smtp_send_article (StatusItem    * item,
 	/* sanity checks */
 	g_return_val_if_fail (item!=NULL, -1);
 	g_return_val_if_fail (sock!=NULL, -1);
-	g_return_val_if_fail (article!=NULL, -1);
+	g_return_val_if_fail (article_is_valid(article), -1);
 
 	status_item_emit_status (item, _("Preparing to Send Mail"));
 
@@ -216,8 +218,11 @@ smtp_send_article (StatusItem    * item,
 		gchar * author = article_get_author_str (article);
 		if (is_nonempty_string (author))
 			from = g_strdup (author);
-		if (from == NULL)
-			from = get_default_author_address ();
+		if (from == NULL) {
+			Identity * id = identity_manager_get_default (ID_MAIL_DEFAULT);
+			if (id != NULL)
+				from = g_strdup (id->author_addr);
+		}
 		if (from == NULL) {
 			status_item_emit_error(item, _("No Sender Specified!"));
 			okay = FALSE;
@@ -331,6 +336,13 @@ smtp_send_article (StatusItem    * item,
 		okay = !smtp_send_line_va (item, sock, "Subject: %s\r\n", tmp);
 		g_free (tmp);
 	}
+	if (okay) {
+		gchar * date = rfc822_date_generate(0);
+		gchar * tmp = g_mime_utils_8bit_header_encode (date);
+		okay = !smtp_send_line_va (item, sock, "Date: %s\r\n", tmp);
+		g_free (tmp);
+		g_free (date);
+	}
 	if (okay && references!=NULL) {
 		gchar * tmp = g_mime_utils_8bit_header_encode (references);
 		okay = !smtp_send_line_va (item, sock, "References: %s\r\n", tmp);
@@ -472,7 +484,7 @@ smtp_send (StatusItem      * item,
 
 	/* sanity checks */
 	g_return_val_if_fail (item!=NULL, -1);
-	g_return_val_if_fail (article!=NULL, -1);
+	g_return_val_if_fail (article_is_valid(article), -1);
 
 	/* connect to the smtp server */
 	sock = smtp_get_connection (item);
@@ -487,3 +499,6 @@ smtp_send (StatusItem      * item,
 	pan_object_unref (PAN_OBJECT(sock));
 	return retval;
 }
+
+
+
