@@ -31,15 +31,15 @@
 
 #include <glib.h>
 
-#include <pan/article.h>
-#include <pan/debug.h>
-#include <pan/log.h>
-#include <pan/pan-glib-extensions.h>
-#include <pan/pan-i18n.h>
+#include <pan/base/article.h>
+#include <pan/base/debug.h>
+#include <pan/base/log.h>
+#include <pan/base/pan-glib-extensions.h>
+#include <pan/base/pan-i18n.h>
 
 #include <pan/nntp.h>
-#include <pan/rule-manager.h>
 #include <pan/task-headers.h>
+#include <pan/rules/rule-manager.h>
 
 /*********************
 **********************  Defines / Enumerated types
@@ -121,7 +121,7 @@ real_task_headers_new (Group                * group,
 	TaskHeaders * item;
 	debug_enter ("real_task_headers_new");
 
-	g_return_val_if_fail (group!=NULL, NULL);
+	g_return_val_if_fail (group_is_valid(group), NULL);
 	g_return_val_if_fail (!group_is_folder(group), NULL);
 	g_return_val_if_fail (group->server!=NULL, NULL);
 	g_return_val_if_fail (dl_type==HEADERS_ALL || dl_type==HEADERS_NEW || dl_type==HEADERS_SAMPLE, NULL);
@@ -278,15 +278,15 @@ task_headers_run (Task* task)
 			/* add the articles to the group */
 			group_mark_new_article_number (task_h->group, task_h->group->article_high);
 			group_add_articles_remove_unused (task_h->group, articles, status);
+			group_expire_articles_not_in_range (task_h->group, low_in_group, high_in_group);
 			group_set_article_range (task_h->group, low_in_group, high_in_group);
-			group_expire_old_articles (task_h->group);
 			group_thread_if_needed (task_h->group, status);
 
 			/* apply the rules to the new headers */
 			rule_manager_process_incoming_articles (articles);
 
 			/* do we need to download these? */
-			if (task_h->download_bodies)
+			if (task_h->download_bodies && articles->len)
 				pan_g_ptr_array_append (task_h->headers_needing_bodies,
 				                        articles->pdata,
 				                        articles->len);
@@ -360,7 +360,7 @@ task_get_header_range (TaskHeaders   * task,
 	/* sanity checks */
 	g_return_val_if_fail (sock!=NULL, TASK_FAIL_HOPELESS);
 	g_return_val_if_fail (server!=NULL, TASK_FAIL_HOPELESS);
-	g_return_val_if_fail (group!=NULL, TASK_FAIL_HOPELESS);
+	g_return_val_if_fail (group_is_valid(group), TASK_FAIL_HOPELESS);
 
 	/* change to the right group */
 	lo_article_in_group = 0;
@@ -391,7 +391,7 @@ task_get_header_range (TaskHeaders   * task,
 
 		lo_article_to_fetch = p_last!=0 ? p_last : lo_article_in_group;
 		hi_article_to_fetch = hi_article_in_group;
-		progress_fmt = _("Getting new Article Header %lu of %lu");
+		progress_fmt = _("Getting new Headers %lu of %lu");
 	}
 	else if (dl_type==HEADERS_SAMPLE)
 	{
@@ -400,13 +400,13 @@ task_get_header_range (TaskHeaders   * task,
 			STATUS_ITEM(task), _("Sampling %d articles"), qty);
 		lo_article_to_fetch = hi_article_in_group - qty;
 		hi_article_to_fetch = hi_article_in_group;
-		progress_fmt = _("Sampling Article Header %lu of %lu");
+		progress_fmt = _("Sampling Headers %lu of %lu");
 	}
 	else if (dl_type==HEADERS_ALL)
 	{
 		lo_article_to_fetch = lo_article_in_group;
 		hi_article_to_fetch = hi_article_in_group;
-		progress_fmt = _("Getting Article Header %lu of %lu");
+		progress_fmt = _("Getting Headers %lu of %lu");
 	}
 	else
 	{
