@@ -25,15 +25,15 @@
 
 #include <config.h>
 
-#include <gnome.h>
+#include <libgnomeui/libgnomeui.h>
 
-#include <pan/debug.h>
-#include <pan/pan-glib-extensions.h>
+#include <pan/base/debug.h>
+#include <pan/base/pan-i18n.h>
+#include <pan/base/pan-glib-extensions.h>
 
 #include <pan/gui.h>
 #include <pan/prefs.h>
 #include <pan/queue.h>
-#include <pan/status-item-view.h>
 #include <pan/task-manager.h>
 #include <pan/util.h>
 
@@ -70,44 +70,12 @@ typedef struct
 }
 ManagerUI;
 
-/*********************
-**********************  Private Function Prototypes
-*********************/
-
-/*********************
-**********************  Constants
-*********************/
-
-/***********
-************  Extern
-***********/
-
-/***********
-************  Public
-***********/
-
-/***********
-************  Private
-***********/
-
 /*********************
 **********************  Variables
 *********************/
 
-/***********
-************  Extern
-***********/
-
 extern GtkTooltips * ttips;
 
-/***********
-************  Public
-***********/
-
-/***********
-************  Private
-***********/
-
 static ManagerUI * manager_ui = NULL;
 
 /*********************
@@ -137,6 +105,95 @@ get_selected_tasks (ManagerUI * ui)
 	return a;
 }
 
+/**
+*** Helper functions for user interactions
+**/
+
+static void
+move_selected_tasks_to_top (ManagerUI * ui)
+{
+	GtkCList * clist;
+	GList * l;
+	gint new_row;
+	debug_enter ("move_selected_tasks_to_top");
+
+	/* sanity clause */
+	g_return_if_fail (ui!=NULL);
+
+	clist = GTK_CLIST(ui->list);
+	new_row = -1;
+	for (l=clist->selection; l!=NULL; l=l->next)
+	{
+		const int old_row = GPOINTER_TO_INT (l->data);
+		Task * const task = gtk_clist_get_row_data (clist, old_row);
+		queue_task_move (task, ++new_row);
+	}
+
+	debug_exit ("move_selected_tasks_to_top");
+}
+
+static void
+move_selected_tasks_to_bottom (ManagerUI * ui)
+{
+	GtkCList * clist;
+	GList * l;
+	gint new_row;
+	debug_enter ("move_selected_tasks_to_bottom");
+
+	/* sanity clause */
+	g_return_if_fail (ui!=NULL);
+
+	clist = GTK_CLIST(ui->list);
+	new_row = 0;
+	for (l=clist->selection; l!=NULL; l=l->next)
+	{
+		const int old_row = GPOINTER_TO_INT (l->data);
+		Task * const task = gtk_clist_get_row_data (clist, old_row);
+		queue_task_move (task, clist->rows - ++new_row);
+	}
+
+	debug_exit ("move_selected_tasks_to_bottom");
+}
+
+static void
+cancel_selected_tasks (ManagerUI * ui)
+{
+	GPtrArray * a;
+        debug_enter ("cancel_selected_tasks");
+
+	/* sanity clause */
+	g_return_if_fail (ui!=NULL);
+
+	a = get_selected_tasks (ui);
+	pan_g_ptr_array_foreach (a, (GFunc)queue_task_remove, NULL);
+	g_ptr_array_free (a, TRUE);
+
+        debug_exit ("cancel_selected_tasks");
+}
+
+static void
+select_task (ManagerUI * ui, gboolean first)
+{
+	GtkCList * clist;
+	gint       row;
+	debug_enter ("select_task");
+
+	/* sanity clause */
+	g_return_if_fail (ui!=NULL);
+
+	clist = GTK_CLIST(ui->list);
+	row = first ? 0 : clist->rows-1;
+	
+	gtk_clist_freeze(clist);
+	gtk_clist_unselect_all (clist);
+	gtk_clist_select_row (clist, row, 0);
+	clist->focus_row = row;
+	gtk_clist_thaw(clist);
+
+	debug_exit("select_task");
+}
+
+
 /**
 ***  User Interactions
 **/
@@ -200,21 +257,9 @@ static void
 top_clicked_cb (GtkButton    * button,
                 gpointer       user_data)
 {
-        ManagerUI * ui;
-	GtkCList * clist;
-	GList * l;
-	gint new_row;
 	debug_enter ("top_clicked_cb");
 
-        ui = (ManagerUI*) user_data;
-	clist = GTK_CLIST(ui->list);
-	new_row = -1;
-	for (l=clist->selection; l!=NULL; l=l->next)
-	{
-		const int old_row = GPOINTER_TO_INT (l->data);
-		Task * const task = gtk_clist_get_row_data (clist, old_row);
-		queue_task_move (task, ++new_row);
-	}
+	move_selected_tasks_to_top ((ManagerUI *) user_data);
 
 	debug_exit ("top_clicked_cb");
 }
@@ -244,21 +289,9 @@ static void
 bottom_clicked_cb (GtkButton   * button,
                    gpointer      user_data)
 {
-        ManagerUI * ui;
-	GtkCList * clist;
-	GList * l;
-	gint new_row;
 	debug_enter ("bottom_clicked_cb");
 
-        ui = (ManagerUI*) user_data;
-	clist = GTK_CLIST(ui->list);
-	new_row = 0;
-	for (l=clist->selection; l!=NULL; l=l->next)
-	{
-		const int old_row = GPOINTER_TO_INT (l->data);
-		Task * const task = gtk_clist_get_row_data (clist, old_row);
-		queue_task_move (task, clist->rows - ++new_row);
-	}
+	move_selected_tasks_to_bottom ((ManagerUI *) user_data);
 
 	debug_exit ("bottom_clicked_cb");
 }
@@ -267,14 +300,9 @@ static void
 cancel_clicked_cb (GtkButton   * button,
                    gpointer      user_data)
 {
-	ManagerUI * ui;
-	GPtrArray * a;
 	debug_enter ("cancel_clicked");
 
-	ui = (ManagerUI*) user_data;
-	a = get_selected_tasks (ui);
-	pan_g_ptr_array_foreach (a, (GFunc)queue_task_remove, NULL);
-	g_ptr_array_free (a, TRUE);
+	cancel_selected_tasks ((ManagerUI *) user_data);
 
 	debug_exit ("cancel_clicked");
 }
@@ -283,18 +311,20 @@ static void
 clear_queue_clicked_cb (GtkButton   * b,
                         gpointer      user_data)
 {
-	GSList * l;
-	GSList * queue;
+	guint i;
+	GPtrArray * tasks;
 	debug_enter ("clear_queue_clicked_cb");
 
-	queue = g_slist_copy ((GSList*)queue_get_tasks());
-	for (l=queue; l!=NULL; l=l->next) {
-		Task * task = TASK(l->data);
+	tasks = queue_get_tasks ();
+	for (i=0; i<tasks->len; ++i) {
+		Task * task = TASK(g_ptr_array_index (tasks,i));
 		if (queue_get_task_status(task)==QUEUE_TASK_STATUS_QUEUED)
 			queue_task_remove(task);
 	}
 
-	g_slist_free (queue);
+	/* cleanup */
+	pan_g_ptr_array_foreach (tasks, (GFunc)pan_object_unref, NULL);
+	g_ptr_array_free (tasks, TRUE);
 
 	debug_exit ("clear_queue_clicked_cb");
 }
@@ -452,10 +482,16 @@ add_task_to_list (ManagerUI     * ui,
 static void
 ui_populate (ManagerUI *ui)
 {
-	int i;
-	const GSList* l = NULL;
-	for (i=0, l=queue_get_tasks(); l!=NULL; l=l->next, ++i)
-		add_task_to_list (ui, l->data, i);
+	guint i;
+	GPtrArray * tasks;
+
+	tasks = queue_get_tasks ();
+	for (i=0; i<tasks->len; ++i)
+		add_task_to_list (ui, TASK(g_ptr_array_index(tasks,i)), i);
+
+	/* cleanup */
+	pan_g_ptr_array_foreach (tasks, (GFunc)pan_object_unref, NULL);
+	g_ptr_array_free (tasks, TRUE);
 }
 
 static void
@@ -481,42 +517,32 @@ queue_row_move_cb (GtkCList   * clist,
 ***  Queue Changes
 **/
 
-static int
-task_added_cb (gpointer call_obj,
-               gpointer call_arg,
-               gpointer client_arg)
+static void
+task_added_cb (gpointer call_obj, gpointer call_arg, gpointer client_arg)
 {
 	Task * task = TASK(call_obj);
 	int index = GPOINTER_TO_INT(call_arg);
 	ManagerUI * ui = (ManagerUI*) client_arg;
 
 	add_task_to_list (ui, task, index);
-
-	return 0;
 }
 
-static int
-task_removed_cb (gpointer call_obj,
-                 gpointer call_arg,
-                 gpointer client_arg)
+static void
+task_removed_cb (gpointer call_obj, gpointer call_arg, gpointer client_arg)
 {
 	Task * task = TASK(call_obj);
 	int index = GPOINTER_TO_INT(call_arg);
 	ManagerUI * ui = (ManagerUI*) client_arg;
 	GtkCList * list = GTK_CLIST(ui->list);
 
-	g_return_val_if_fail (gtk_clist_get_row_data(list,index)==task, 0);
+	g_return_if_fail (gtk_clist_get_row_data(list,index)==task);
 	pan_lock ();
 	gtk_clist_remove (list, index);
 	pan_unlock ();
-
-	return 0;
 }
 
-static int
-task_moved_cb (gpointer call_obj,
-               gpointer call_arg,
-               gpointer client_arg)
+static void
+task_moved_cb (gpointer call_obj, gpointer call_arg, gpointer client_arg)
 {
 	ManagerUI * ui = (ManagerUI*) client_arg;
 	int old_row = ((int*)call_arg)[0];
@@ -527,22 +553,70 @@ task_moved_cb (gpointer call_obj,
 	gtk_clist_row_move (GTK_CLIST(ui->list), old_row, new_row);
 	ui->moved_signal_id = gtk_signal_connect (GTK_OBJECT(ui->list), "row_move", queue_row_move_cb, ui);
 	pan_unlock ();
+}
 
-	return 0;
+static gboolean
+gui_key_press_cb (GtkWidget      * widget,
+                  GdkEventKey    * event,
+                  gpointer         data)
+{
+	ManagerUI * ui;
+	GtkCList  * clist;
+	debug_enter ("gui_key_press_cb");
+
+	ui    = (ManagerUI*) data;
+	clist = GTK_CLIST (ui->list);
+
+        switch (event->keyval)
+        {
+                case GDK_Delete:
+			cancel_selected_tasks (ui);
+			break;
+                case GDK_Home:
+                        if (event->state & GDK_SHIFT_MASK)
+			{
+				move_selected_tasks_to_top (ui);
+			}
+			else
+			{
+				select_task (ui, TRUE);
+			}
+			break;
+                case GDK_End:
+                        if (event->state & GDK_SHIFT_MASK)
+			{
+				move_selected_tasks_to_bottom (ui);
+			}
+			else
+			{
+				select_task (ui, FALSE);
+			}
+			break;
+                default:
+			break;
+        }
+
+	debug_exit ("gui_key_press_cb");
+        return TRUE;
 }
 
 static void
 update_titlebar_lock_unconditional (ManagerUI * ui)
 {
 	int running, queued, failed;
-	const GSList* l;
+	guint i;
+	GPtrArray * tasks;
 	gchar* pch;
 
 	g_return_if_fail (ui!=NULL);
 
+	/* get information about the tasks */
 	running = queued = failed = 0;
-	for (l=queue_get_tasks(); l!=NULL; l=l->next) {
-		QueueTaskStatus status = queue_get_task_status (TASK(l->data));
+	tasks = queue_get_tasks ();
+	for (i=0; i<tasks->len; ++i) {
+		Task * t = TASK(g_ptr_array_index(tasks,i));
+		QueueTaskStatus status = queue_get_task_status (t);
+	
 		if (status==QUEUE_TASK_STATUS_QUEUED)
 			++queued;
 		else if (status==QUEUE_TASK_STATUS_RUNNING)
@@ -551,16 +625,23 @@ update_titlebar_lock_unconditional (ManagerUI * ui)
 			++failed;
 	}
 
+	/* create a titlebar */
 	if (failed)
 		pch = g_strdup_printf (_("Pan %s Task Manager (%d Queued, %d Running, %d Failed)"), VERSION, queued, running, failed);
 	else if (queued || running)
 		pch = g_strdup_printf (_("Pan %s Task Manager (%d Queued, %d Running)"), VERSION, queued, running);
 	else 
 		pch = g_strdup_printf (_("Pan %s Task Manager"), VERSION);
+
+	/* show the titlebar */
 	pan_lock_unconditional ();
 	gtk_window_set_title (GTK_WINDOW(ui->window), pch);
 	pan_unlock_unconditional ();
 	g_free (pch);
+
+	/* cleanup */
+	pan_g_ptr_array_foreach (tasks, (GFunc)pan_object_unref, NULL);
+	g_ptr_array_free (tasks, TRUE);
 }
 
 static int
@@ -796,13 +877,10 @@ task_manager_build_ui (GtkWidget * window)
 ***  Titlebar dirty
 **/
 
-static int
-update_titlebar_cb (gpointer call_obj,
-                    gpointer call_arg,
-                    gpointer user_data)
+static void
+update_titlebar_cb (gpointer call_obj, gpointer call_arg, gpointer user_data)
 {
 	((ManagerUI*)(user_data))->titlebar_needs_refresh = TRUE;
-	return 0;
 }
 static void
 remove_update_titlebar_callbacks_cb (GtkObject * o, gpointer user_data)
@@ -841,6 +919,8 @@ task_manager_spawn (void)
 	                    GTK_SIGNAL_FUNC(window_delete_event_cb), manager_ui);
 	gtk_signal_connect (GTK_OBJECT(manager_ui->window), "destroy",
 	                    GTK_SIGNAL_FUNC(remove_update_titlebar_callbacks_cb), manager_ui);
+	gtk_signal_connect (GTK_OBJECT (manager_ui->window), "key_press_event",
+	                    GTK_SIGNAL_FUNC(gui_key_press_cb), manager_ui);
 	gtk_window_set_title (GTK_WINDOW(manager_ui->window), _("Pan - Task Manager"));
 	gtk_container_add (GTK_CONTAINER(manager_ui->window), manager_ui->top);
 
